open Printf
open Sequence

(** Default character for alignment gaps *)
let defgapchar = '-'
(** Default character for alignment comparison gaps *)
let compgap = '.'

type align_seq = int option array

(** Create a string, possibly gapped, for the alignment of [x]
    (numeric) according to [xmap], the numeric indices from
    alignment. Use translation table [trans] *)
let alignment_string ?(gapchar=defgapchar) trans x xmap allen =
  let xal = String.create allen in
  for i=0 to allen-1 do
    match xmap.(i) with
      None -> xal.[i] <- gapchar
    | Some(j) -> 
	let c = try trans.[x.(j)] with Invalid_argument(_) -> 
	  failwith (sprintf "Problem with translation table, char %d out of bounds" x.(j)) in
	xal.[i] <- c
  done;
  xal;;


(** Types of output sequences for standard alignment output *)
type outseq = 
    AAsequence 
  | SSsequence 
  | SSpred2true 
  | IndexSeq of string

(** Figure out what part of the sequence and which translation table to use for output *)
let get_outseq_array ?(trans_table="01") xs ys seqstyle =
  match seqstyle with
  | AAsequence ->    (xs.seq, ys.seq, Sequence.totaaorder)
  | SSsequence ->    (xs.ss, ys.ss, Sequence.ssorder)
  | SSpred2true ->   (xs.sspred, ys.ss, Sequence.ssorder)
  | IndexSeq(idxname) -> 
      let xindex = Sequence.get_index xs idxname in
      let yindex = Sequence.get_index ys idxname in
      (xindex, yindex, trans_table);;


(** Computes the percent identity of aligned sequences, returns both
    exact and inexact (positive blosum scores) identities. *)
let seq_identity x y xmap ymap alen =
  let s,r,n = ref 0, ref 0, ref 0 in
  for i=0 to alen-1 do
    match xmap.(i),ymap.(i) with
      Some(xi),Some(yi) -> 
	incr n;
	if x.(xi) = y.(yi) then incr s;	(* Exact identity *)
	if Matrices.blosum62.(x.(xi)).(y.(yi)) > 0.0 then incr r; (* Inexact identity *)
    | _ -> ();
  done;
  let nn = float !n in
  (float !s) /. nn, (float !r) /. nn;;

(** Create a string that reflects comparison between two
    aligments. [xmap] is the original aligment and [cmap] is how the
    two alignments are aligned. *)
let alcomp_string ?(gapchar=defgapchar) x xmap cmap allen =
  let xal = String.create allen in
  for i=0 to allen-1 do
    xal.[i] <- match cmap.(i) with
      None -> compgap
    | Some(j) -> match xmap.(j) with
	None -> gapchar
      | Some(k) -> Sequence.i2c x.(k)
  done;
  xal;;
	
(** Create a string that indicates matching positions in an alignment
    according to the maps. *)
let match_string x y xmap ymap allen =
  let mal = String.create allen in
  for i=0 to allen-1 do
    mal.[i] <- match xmap.(i),ymap.(i) with
      Some(i),Some(j) when x.(i)=y.(j) -> '|'
    | _ -> ' '
  done;
  mal;;

(** Type of output style for alignments *)
type outstyle =
    Standard	 (** Easy to read/parse format, allow different translation table *)
  | PIR		 (** PIR alignment output style, input to Modeller *)
  | C7pred	 (** Style of c7pred output for alignments *)

(** Alignment output, multiple formats possible *)
let print_alignment style 
    ?(ids=("seq1","seq2")) ?(fout=None) (x,y,trans) xmap ymap score allen =
  let f = Util.opt_open fout in
  begin match style with
    Standard ->
      let xal = alignment_string trans x xmap allen in
      let yal = alignment_string trans y ymap allen in
      let eid,iid = seq_identity x y xmap ymap allen in
      let xid,yid = ids in
      fprintf f "%6s to %6s Score: %6.1f Len1: %4d Len2: %4d ALen: %4d EIdent: %4.2f IIdent: %4.2f\n%s\n%s\n" 
	xid yid score (Array.length x) (Array.length y) allen eid iid xal yal;
  | PIR ->
      let xal = Util.wordwrap ~hard:true 75 (alignment_string trans x xmap allen) in
      let yal = Util.wordwrap ~hard:true 75 (alignment_string trans y ymap allen) in
      let rec scan next a i = match a.(i) with 
	  Some(j) -> j
	| _ -> scan next a (next i) in
      let xstart = scan ((+) 1) xmap 0 in
      let xend = scan ((+) (-1)) xmap (allen-1) in
      let ystart = scan ((+) 1) ymap 0 in
      let yend = scan ((+) (-1)) ymap (allen-1) in
      let xid,yid = ids in
      fprintf f ">P1;%s\nsequence:%s:%d:@:%d:@:\n%s*\n\n" xid xid (xstart+1) (xend+1) xal;
      fprintf f ">P1;%s\nstructure:%s:%d:@:%d:@:\n%s*\n"  yid yid (ystart+1) (yend+1) yal;
  | C7pred ->
      let xal = alignment_string trans x xmap allen in
      let yal = alignment_string trans y ymap allen in
      let _,yid = ids in
      fprintf f "PARENT %s\n" yid;
      for i=0 to allen-1 do
	match xmap.(i),ymap.(i) with
	  Some(xi),Some(yj) -> 
	    fprintf f " %c %4d %c %d\n" xal.[i] (xi+1) yal.[i] (yj+1);
	| _ -> ()
      done;
      fprintf f "TER\n";
  end;
  Util.opt_close fout f;;

(** PIR alignment output style, input to Modeller *)
let output_pir_alignment ?(ids=("seq1","seq2")) ?(fout=None) (trans,x,y) xmap ymap allen =
  let f = Util.opt_open fout in
  let xal = Util.wordwrap ~hard:true 75 (alignment_string trans x xmap allen) in
  let yal = Util.wordwrap ~hard:true 75 (alignment_string trans y ymap allen) in
  let rec scan next a i = match a.(i) with 
      Some(j) -> j
    | _ -> scan next a (next i) in
  let xstart = scan ((+) 1) xmap 0 in
  let xend = scan ((+) (-1)) xmap (allen-1) in
  let ystart = scan ((+) 1) ymap 0 in
  let yend = scan ((+) (-1)) ymap (allen-1) in
  let xid,yid = ids in
  fprintf f ">P1;%s\nsequence:%s:%d:@:%d:@:\n%s*\n\n"
    xid xid (xstart+1) (xend+1) xal;
  fprintf f ">P1;%s\nstructure:%s:%d:@:%d:@:\n%s*\n"
    yid yid (ystart+1) (yend+1) yal;
  Util.opt_close fout f;;
  
(*   | Compare_aligns(nseq1,nseq2,xmap1,ymap1,xmap2,ymap2) -> *)
      
let print_compare_alignment ?(ids=("seq1","seq2")) ?(fout=None) 
    xs ys xmap ymap score allen nseq1 nseq2 xmap1 ymap1 xmap2 ymap2 =
  let f = Util.opt_open fout in
  let x,y = xs.seq, ys.seq in
  let m,n = Array.length xmap1, Array.length xmap2 in
  assert(m=Array.length ymap1 && n=Array.length ymap2);
  let eid = score /. float (min m n) in
  let nid = truncate score in
  let len_ratio = (float (min m n)) /. (float (max m n)) in
  let ns = eid *. len_ratio in
  let xid,yid = ids in
  fprintf f "%6s to %6s Len1: %4d Len2 %4d Nident: %4d %%Ident: %4.2f Score: %4.2f\n"
    xid yid m n nid eid ns;
  let xal1 = alcomp_string x xmap1 xmap allen in
  let yal1 = alcomp_string y ymap1 xmap allen in
  let xal2 = alcomp_string x xmap2 ymap allen in
  let yal2 = alcomp_string y ymap2 ymap allen in
  let mal = match_string nseq1 nseq2 xmap ymap allen in
  fprintf f "%s\n%s\n%s\n%s\n%s\n" xal1 yal1 mal xal2 yal2;      
  Util.opt_close fout f;;

(** Convert and alignment character string into a numeric sequence and
    a numeric alignment array *)
let alignstr2num trans s =
  let n = String.length s in
  let seq = Queue.create () in
  let al = Array.make n None in
  let idx = ref 0 in
  for i=0 to n-1 do
    if s.[i]<>defgapchar then begin
      al.(i) <- Some(!idx); 
      incr idx;
      Queue.add trans.(int_of_char s.[i]) seq;
    end;
  done;
  (Util.q2array seq), al;;

open Scanf

(** Read a MAMMOTH output file and extract the sequence alignment
    within it *)
let read_mammoth ?(gapchar='-') fname =
  let f = open_in fname in
  while (input_line f) <> "  Final Structural Alignment " do (); done;
  for i=1 to 3 do ignore(input_line f); done;
  let x,y = Buffer.create 500, Buffer.create 500 in
  let line = ref (input_line f) in
  while (String.sub !line 0 10) = "Prediction" do
    let b = Scanning.from_string !line in
    bscanf b "%s " ignore;
    while not (Scanning.end_of_input b) do
      bscanf b " %s " (fun s-> Buffer.add_string x s)
    done;
    for i=1 to 3 do ignore(input_line f); done;
    line := input_line f;
    let b = Scanning.from_string !line in
    bscanf b "%s " ignore;
    while not (Scanning.end_of_input b) do
      bscanf b " %s " (fun s-> Buffer.add_string y s)
    done;
    for i=1 to 3 do ignore(input_line f); done;
    line := input_line f;
  done;
  let x,y = Buffer.contents x, Buffer.contents y in
  for i=0 to (String.length x)-1 do
    if x.[i]='.' then x.[i] <- gapchar;
    if y.[i]='.' then y.[i] <- gapchar;
  done;
  close_in f;
  x,y;;

(** Read a PIR alignment *)
let read_pir fname =
  let f = open_in fname in
  let b = Buffer.create 500 in
  ignore(input_line f); ignore(input_line f);
  let s = ref (input_line f) in
  while String.length !s > 0 do
    Buffer.add_string b !s;
    s := input_line f;
  done;
  let x = Buffer.sub b 0 ((Buffer.length b)-1) in (* Chop final * char *)
  Buffer.clear b;
  ignore(input_line f); ignore(input_line f);
  s := input_line f;
  begin try 
    while true do
      Buffer.add_string b !s;
      s := input_line f;
    done 
  with End_of_file -> (); end;
  close_in f;
  let y = Buffer.sub b 0 ((Buffer.length b)-1) in (* Chop final * char *)
  x,y;;  
  
(** Create a numeric array and alignment array from a string [s]
    according to translation table [trans], stop at [alen] *)
let alignstring2seq ?(gapchar='-') trans s alen =
  let al = Array.make alen None in
  let seq = Queue.create () in
  let idx = ref 0 in
  for i=0 to alen-1 do
    if s.[i]<>gapchar then begin
      al.(i) <- Some(!idx); 
      incr idx;
      Queue.add trans.(int_of_char s.[i]) seq;
    end;
  done;
  (Util.q2array seq), al;;

(* (\** Extract numeric sequence from alignment array *\) *)
(* let alignarray2numseq a = *)
(*   let q = Queue.create () in *)
(*   let op x = match x with *)
(*       None -> () *)
(*     | Some(i) -> Queue.add i q  *)
(*   in *)
(*   Array.iter op a; *)
(*   Util.q2array q;; *)

(** Read in a c7pred style alignment.  This version punts on missing
    characters and relies on later synchronization to fill in gapped
    sections which are not present in the alignment. *)
let read_c7pred_align fname =
  let n = (Util.count_lines fname)-2 in
  let xq,yq = (Queue.create ()),(Queue.create ()) in
  let f = Scanning.from_file fname in
  bscanf f " %_s %_s " ();
  let lastx = ref 0 in
  let lasty = ref 0 in
  for i=0 to n-1 do
    let read xc xi yc yj =
      if i=0 then (lastx:=xi-1; lasty:=yj-1);
      begin match (xi - !lastx),(yj - !lasty) with
	1,1 -> ()
      | 1,m -> for j=1 to m-1 do Queue.add '-' xq; Queue.add 'X' yq; done;
      | m,1 -> for j=1 to m-1 do Queue.add 'X' xq; Queue.add '-' yq; done;
      | _ -> failwith (sprintf "Error reading %s: aligned gaps?" fname)
      end;
      Queue.add xc xq; Queue.add yc yq;
      lastx := xi; lasty := yj;
    in
    bscanf f " %c %d %c %d " read;
  done;
  bscanf f " %_s " ();			(* Finish last TER field *)
  let xals = Util.chars2string (Util.q2list xq) in
  let yals = Util.chars2string (Util.q2list yq) in
  xals,yals


(** Read an alignment file, return numeric seqeunce of both entries
    and numeric alignment array, limited support so far *)
let read_alignment ?(gapchar='-') ?(trans=aacodes1) style fname =
  let xals,yals = match style with 
      "standard" -> 
	let f = open_in fname in
	ignore(input_line f);
	let x = input_line f in
	let y = input_line f in
	close_in f;
	x,y
    | "mammoth" -> read_mammoth fname
    | "pir" -> read_pir fname
    | "c7pred" -> read_c7pred_align fname
    | _ -> failwith (sprintf "Unknown alignment type %s for file %s" style fname)
  in
  let alen = String.length xals in
  assert(alen = String.length yals);
  let x, xal = alignstring2seq ~gapchar trans xals alen in
  let y, yal = alignstring2seq ~gapchar trans yals alen in
  (x, y, xal, yal);;

(*   let x,y = Buffer.create 200, Buffer.create 200 in *)
(*   let xi,yi = ref 0, ref 0 in *)
(*   let xmap,ymap = Queue.create (), Queue.create () in *)
(*   for i=0 to alen-1 do *)
(*     if xal.[i] = gapchar then *)
(*       Queue.add None xmap *)
(*     else begin *)
(*       Buffer.add_char x xal.[i]; *)
(*       Queue.add (Some !xi) xmap; *)
(*       incr xi; *)
(*     end; *)
(*     if yal.[i] = gapchar then *)
(*       Queue.add None ymap *)
(*     else begin *)
(*       Buffer.add_char y yal.[i]; *)
(*       Queue.add (Some !yi) ymap; *)
(*       incr yi; *)
(*     end; *)
(*   done; *)
(*   (aa2idx (Buffer.contents x)),(aa2idx (Buffer.contents y)), *)
(*   (Util.q2array xmap),(Util.q2array ymap);; *)

open Align

(** Evaluate an alignment, set up scoring, score, and return
    data. This would be more efficient if if we could use the single
    scoring variants but it doesn't fit well in the framework of
    aligntool right now. *)
let eval_alignment x y xmap ymap allen =
  let scores = Array.make_matrix x.seqlen y.seqlen 0.0 in
  Scoring.setup_scoring_mat scores x y;
  let score = score_alignment xmap ymap scores 0 allen in
  score,scores;;

(** Split the filename of a alignment file, try to come up with a
    reasonable label for each sequence *)
let split_filename fname = 
  let f = Util.basename fname in
  let split c s = 
    let idx = String.index s c in
    (String.sub s 0 idx),(String.sub s (idx+1) ((String.length s)-idx-1))
  in
  if String.contains f '_' 
  then split '_' f
  else if String.contains f '-' 
  then split '-' f 
  else if String.contains f '.' 
  then split '.' f
  else f,f;;

open Scoring

(** Find a sequence correspondence between [x] and [y] according to
    matching.*)
let find_seq_correspondence ?(fout=None) x y =
  let params = Align.set_params 
    ~scores:[seq_match_scores ~scale:1.0]
    ~al:Align.GlobalEF ~go:1.0 ~ge:0.0 () in
  let score,allen,map1,map2,_ = allocate_align x y in
  Align.reset_params params;
  (* Create a correspondence between x and y *)
  let corr = Array.make x.seqlen None in
  for i=0 to allen-1 do
    match map1.(i),map2.(i) with
      Some(j),Some(k) -> corr.(j) <- Some(k)
    | Some(j),None -> corr.(j) <- None	(* Can potentially lead to gap-gap alignment *)
    | None,Some(_) -> ()
    | None,None -> failwith "Aligned gaps in sequence synchronization"
  done;
  let seqtrans = (x.seq,y.seq,Sequence.totaaorder) in
  begin match fout with
    None -> ()
  | Some(_) -> print_alignment Standard ~fout seqtrans map1 map2 score allen 
  end;
  corr;;

(** Adjust xmap so that it reflects the correspondence to another sequence *)
let fixmap corr xmap =
  let n= Array.length xmap in
  let ymap = Array.make n None in
  for i=0 to n-1 do
    match xmap.(i) with
      Some(j) -> ymap.(i) <- corr.(j)
    | None -> ymap.(i) <- None
  done;
  ymap;;


(** Read a an alignment and attempt to correspond the two sequences
    within to given independent sequences. If given sequences have no
    [seq] field, then just use what is given in the alignment. Throws
    exceptions if no correspondence is found. Return the synchronized 
    sequences and their alignment maps. *)
let read_and_sync_align2seq ?(corrout=false) atype fname seq1 seq2 =
  let x,y,xmap,ymap = read_alignment atype fname in
  let dosync z zmap base i =
    match base.seq with
      [||] -> {base with seqlen=Array.length z; seq=z},zmap
    | _ ->
	let fout = 
	  if corrout then Some(fname^".cor."^(string_of_int i)) else None in
	let zs = {(new_seq ()) with seqlen=Array.length z; seq=z} in
	let corr = find_seq_correspondence ~fout zs base in
	base,(fixmap corr zmap)
  in
  let seq1new,seq1map = 
    try dosync x xmap seq1 1
    with Failure(s) -> failwith (fname ^ "Seq 1: " ^s) in
  let seq2new,seq2map = 
    try dosync y ymap seq2 2
    with Failure(s) -> failwith (fname ^ "Seq 2: " ^s) in
  seq1new,seq2new,seq1map,seq2map;;

(** Convert an alignment (two maps) into a numeric sequence based on
    the lengths of [m,n] of the participating sequences. *)
let alignment2numeric_seq m n xmap ymap =
  let len = Array.length xmap in
  let nseq = Array.make len 0 in
  for k=0 to len-1 do
    nseq.(k) <- match xmap.(k),ymap.(k) with
    | Some(i),None -> i
    | None,Some(j) -> m+j
    | Some(i),Some(j) -> m+n + (i*n)+j
    | None,None -> failwith "Aligned gaps in alignment2numeric_seq"
  done;
  nseq;;

(** A record in a sequence alignment database *)
type alignDBentry = {
  xid: string;
  yid: string;
  score:float;
  xlen:int;
  ylen:int;
  alen:int;
  eident:float;
  iident:float;
  xals:string;
  yals:string;
(*   xnseq:int array; *)
(*   ynseq:int array; *)
(*   xal:int option array; *)
(*   yal:int option array; *)
}

(*     let xnseq, xal = alignstring2seq aacodes1 xals alen in *)
(*     let ynseq, yal = alignstring2seq aacodes1 yals alen in *)
(*       xnseq=xnseq; ynseq=ynseq; xal=xal; yal=yal; } *)

(** Read a single standard alignment from a scanning buffer and return
    its contents as an [alignDBentry] *)
let read_one_stand_align f =
  let op xid yid score xlen ylen alen eident iident xals yals =
    { xid=xid; yid=yid; score=score; 
      xlen=xlen; ylen=ylen; alen=alen; eident=eident; iident=iident; 
      xals=xals; yals=yals; }
  in
  bscanf f " %s %_s %s %_s %f %_s %d %_s %d %_s %d %_s %f %_s %f %s %s " op;;

(** Read a file containing multiple standard alignments, create a
    database of them and return. No sequence synchronization is done
    so they better be global alignments. *)
let read_standard_alignmentDB fname =
  printf "Reading alignment DB from %s\n" fname; flush stdout;
  let nDB = 
    let n = Util.count_lines fname in
    n / 3 in
  let db = Hashtbl.create nDB in
  let f = Scanning.from_file fname in
  try 
    while not (Scanning.end_of_input f) do
      let entry = read_one_stand_align f in
      Hashtbl.replace db (entry.xid, entry.yid) entry;
    done;
    db
  with e -> 
    eprintf "Problems with alignment DB file %s\n" fname; 
    raise e;;

(** Perform some cursory checks of an alignDB against the seqDB *)
let check_alignDB alignDB seqDB =
  let rm_aligns = Queue.create () in
  let nongapchars xs =
    let n = String.length xs in
    let gap = '-' in
    let ng = ref 0 in
    for i=0 to n-1 do if xs.[i]=gap then incr ng; done;
    !ng
  in
  let op key entry =
    let xid, yid = entry.xid, entry.yid in
    if Hashtbl.mem seqDB xid && Hashtbl.mem seqDB yid then begin
      let seq = Hashtbl.find seqDB xid in
      let allen = nongapchars entry.xals in
      if seq.seqlen < allen then begin
	eprintf "Warning: seqlen mismatch on %s %s for %s: %d to %d\n" xid yid xid allen seq.seqlen;
	Queue.add key rm_aligns
      end;
      let seq = Hashtbl.find seqDB yid in
      let allen = nongapchars entry.yals in
      if seq.seqlen < allen then begin
	eprintf "Warning: seqlen mismatch on %s %s for %s: %d to %d\n" xid yid yid allen seq.seqlen;
	Queue.add key rm_aligns
      end
    end
    else begin
      eprintf "Warning: seqDB does not contain one of %s %s\n" entry.xid entry.yid;
      Queue.add key rm_aligns;
    end
  in
  Hashtbl.iter op alignDB;
  Queue.iter (fun k-> Hashtbl.remove alignDB k) rm_aligns;
  Queue.length rm_aligns;;
  
(* (\** Adjust xmap so that it reflects the correspondence to another sequence *\) *)
(* let fixmap_inplace corr xmap = *)
(*   let n= Array.length xmap in *)
(*   for i=0 to n-1 do *)
(*     match xmap.(i) with *)
(*       Some(j) -> xmap.(i) <- corr.(j) *)
(*     | _ -> () *)
(*   done;; *)

(* let sync_alignDB seqDB alignDB = *)
(*   let badq = Queue.create () in *)
(*   let op key entry = *)
(*     if Hashtbl.mem seqDB entry.xid && Hashtbl.mem seqDB entry.yid then begin *)
(*       try  *)
(* 	let base = Hashtbl.find seqDB entry.xid in *)
(* 	let xs = {empty_seq with seqlen=Array.length entry.xnseq; seq=entry.xnseq} in *)
(* 	let corr = find_seq_correspondence xs base in *)
(* 	fixmap_inplace corr entry.xal; *)
(* 	let base = Hashtbl.find seqDB entry.yid in *)
(* 	let ys = {empty_seq with seqlen=Array.length entry.ynseq; seq=entry.ynseq} in *)
(* 	let corr = find_seq_correspondence ys base in *)
(* 	fixmap_inplace corr entry.yal; *)
(*       with Invalid_argument(s) -> *)
(* 	eprintf "Warning: skipping alignDB entry %s %s : %s.  skipping..\n" entry.xid entry.yid s; *)
(* 	Queue.add key badq *)
(*     end *)
(*     else begin *)
(*       eprintf "Warning: aligDB does not contain %s %s\n" entry.xid entry.yid; *)
(*       Queue.add key badq *)
(*     end *)
(*   in *)
(*   Hashtbl.iter op alignDB; *)
(*   Queue.iter (fun k-> Hashtbl.remove alignDB k) badq; *)
(*   Queue.length badq;; *)


  
