(** Never filter out, always count a db entry regardless of gaps
    around it *)
let never_cnt_filt _ _ _ _ = true;;

(** Global control of window width  *)
let cnt_filt_width = ref 2
(** Global control of window tolerance *)
let cnt_filt_tol = ref 0.2


(** Count only if the alignment [al] is relatively ungapped: percent
    of gap positions in window of [width] is smaller than [tol] *)
let gen_window_cnt_filt al alen idx =
  let width = !cnt_filt_width in
  let tol = !cnt_filt_tol in
  let ngaps = ref 0 in
  let npos = ref 0 in
  for i=idx-width to idx+width do
    if 0<=i && i<alen then begin
      incr npos;
      match al.(i) with
	None -> incr ngaps
      | _ -> ()
    end;
  done;
  ((float !ngaps) /. (float !npos)) < tol;;

(** Window filter on db entry *)
let db_window_cnt_filt _ dal alen idx =
  gen_window_cnt_filt dal alen idx;;

(** Window filter on query *)
let query_window_cnt_filt qal _ alen idx =
  gen_window_cnt_filt qal alen idx;;

(** Window filter on both query and db entry *)
let both_window_cnt_filt qal dal alen idx =
  (gen_window_cnt_filt qal alen idx)
  && (gen_window_cnt_filt dal alen idx)
  
(** Window filter on either query and db entry *)
let either_window_cnt_filt qal dal alen idx =
  (gen_window_cnt_filt qal alen idx)
  || (gen_window_cnt_filt dal alen idx)

(** Add to column counts in [counts] according to aligned positions in
    [xal,yal] and [yindex], counts assumed to be same size as x's
    sequence *)
let count_columns cnt_filt counts yindex xal yal alen =
  for i=0 to alen-1 do
    let countme = cnt_filt xal yal alen i in
    match xal.(i),yal.(i),countme with
      Some(j),Some(k),true -> counts.(j) <- counts.(j) + yindex.(k)
    | _ -> ()
  done;;

open Sequence

(** Type of entries in an MSA, type generalized to allow indexes or SS
    chars eventually *)
type 'a msa_entry =
    Nothing
  | Gap
  | Count of 'a
  | Skip of 'a

open Query_align

(** Create a pseudo-MSA for column counting *)
let colcount_msa cnt_filt seq idxname hits seqDB =
  let m = seq.seqlen in
  let n = Array.length hits in
  let msa = Array.make_matrix m n Nothing in
  for ii=0 to n-1 do
    let _,hit = hits.(ii) in (* Parse out best alignment *)
    let dseq = Hashtbl.find seqDB hit.dbkey in
    let dindex = Sequence.get_index ~key:hit.dbkey dseq idxname in
    for jj=0 to hit.alen-1 do
      let countme = cnt_filt hit.qal hit.dal hit.alen jj in
      match hit.qal.(jj), hit.dal.(jj), countme with
	Some(j),Some(k),true  -> msa.(j).(ii) <- Count dindex.(k)
      | Some(j),Some(k),false -> msa.(j).(ii) <- Skip dindex.(k)
      | Some(j),None,_ -> msa.(j).(ii) <- Gap
      | _ -> ()
    done;
  done;
  msa;;


(** Compute absolute threshold (count/width), threshold without
    ignored columns, and threshold without ignored+gap columns.  *)
let colcount_thresh wgts truek msa =
  let m = Array.length msa in
  let n = truek in
  if Array.length msa.(0) < truek then
    failwith "Requested dimension of threshold counts larger than msa";
  let abst = Array.make m 0.0 in
  let ignt = Array.make m 0.0 in
  let gapt = Array.make m 0.0 in
  let nn = float n in
  for i=0 to m-1 do
    let cnts, igns, gaps = ref 0.0, ref 0, ref 0 in
    for j=0 to n-1 do
      match msa.(i).(j) with
	Nothing -> incr igns
      | Gap -> incr gaps
      | Count x -> cnts := !cnts +. ((float x) *. wgts.(i).(j))
      | Skip _ -> ()
    done;
    let fcnts, figns, fgaps = !cnts, float !igns, float !gaps in
    abst.(i) <- fcnts;
    ignt.(i) <- if !igns = n then 0.0 else fcnts *. (nn /. (nn -. figns));
    gapt.(i) <- if !igns + !gaps = n then 0.0 else fcnts *. (nn /. (nn -. figns -. fgaps));
  done;
  abst, ignt, gapt;;

open Printf

(** Generic writing of MSAs for column counting *)
let write_out_msa conv qfeat msa fname =
  let f = open_out fname in
  let m = Array.length msa in
  let n = Array.length msa.(0) in
  for i=0 to m-1 do
    fprintf f "%s\t" (conv qfeat.(i));
    for j=0 to n-1 do
      let s = match msa.(i).(j) with
          Nothing -> "."
	| Gap -> "-"
	| Count x -> (conv x)
	| Skip x -> "'"^(conv x)
      in
      fprintf f "%2s " s;
    done;
    fprintf f "\n";
  done;
  close_out f;;
      
(** Populate MSA with SASAs of hits rather than labels *)
let sasa_msa cnt_filt seq _ hits seqDB =
  let m = seq.seqlen in
  let n = Array.length hits in
  let msa = Array.make_matrix m n Nothing in
  for ii=0 to n-1 do
    let _,hit = hits.(ii) in (* Parse out best alignment *)
    let dseq = Hashtbl.find seqDB hit.dbkey in
    let sasa = dseq.sasa in
    for jj=0 to hit.alen-1 do
      let countme = cnt_filt hit.qal hit.dal hit.alen jj in
      match hit.qal.(jj), hit.dal.(jj), countme with
	Some(j),Some(k),true  -> msa.(j).(ii) <- Count sasa.(k)
      | Some(j),Some(k),false -> msa.(j).(ii) <- Skip sasa.(k)
      | Some(j),None,_ -> msa.(j).(ii) <- Gap
      | _ -> ()
    done;
  done;
  msa;;

open Util

(** Write SASA consistency for each query position *)
let write_sasa_consistency msa fname =
  let f = open_out fname in
  let m = Array.length msa in
  let n = Array.length msa.(0) in
  for i=0 to m-1 do
    let avg = Avgstd.make () in
    for j=0 to n-1 do
      match msa.(i).(j) with
        Nothing | Gap | Skip _ -> ()
      | Count x -> Avgstd.update avg x;
    done;
    let mean,std,cnt = Avgstd.finalize avg in
    fprintf f "%f %f %f\n" cnt mean std;
  done;
  close_out f;;


(** Write GK's SASA consistency for each query position *)
let write_gksasa_consistency msa fname =
  let f = open_out fname in
  let m = Array.length msa in
  let burvec,expvec = (Array.make m 0.0),(Array.make m 0.0) in
  for i=0 to m-1 do
    let f (bur,tot) y = 
      match y with
        Nothing | Gap | Skip _ -> (bur,tot)
      | Count x when x<= 0.06 -> (bur+1,tot+1)
      | Count _ -> (bur,tot+1)
    in
    let bur,tot = Array.fold_left f (0,0) msa.(i) in
    if tot > 0 then begin
      burvec.(i) <- ((float bur) /. (float tot));
      expvec.(i) <- 1.0 -. burvec.(i);
    end else begin
      burvec.(i) <- 0.0;
      expvec.(i) <- 0.0;
    end;
  done;
  let cons = (dotp burvec expvec) /. (norm2 burvec) /. (norm2 expvec) in
  fprintf f "%f\n" cons;
  close_out f;;

(** Shift [scores] up to [newmin] *)
let shift_to_min newmin scores =
  let minscr = Array.fold_left min min_float scores in
  let shift = if minscr < newmin then newmin -. minscr else 0.0 in
  for i=0 to (Array.length scores)-1 do
    scores.(i) <- scores.(i) +. shift
  done;;

(** Divide each element of [scores] by the sum of [scores] *)
let normalize scores = 
  let sum = Util.array_sumf scores in
  for i=0 to (Array.length scores)-1 do
    scores.(i) <- scores.(i) /. sum
  done;;

(** Weights are the same for each sequence, based on original
    alignment score. Scores are shifted to one if any negative scores
    are present and normalized to get the weights. Note that the
    'matrix' returned should *not* be modified as the row pointers all
    point to the same row *)
let global_msa_weights wgts qseq best =
  let m, n = qseq.seqlen, Array.length best in
  let scores = Array.map (fun (s,_)->s) best in
  shift_to_min 1.0 scores;
  normalize scores;
  for i=0 to m-1 do
    for j=0 to n-1 do
      wgts.(i).(j) <- scores.(j)
    done;
  done;;
  
(** Weights vary by position and sequence. A local sequence alignment
    score is computed, shifted, and normalized to determine the
    weight. [width] is the number *alignment* of positions around the
    central residue for the computation. *)
let local_msa_weights wgts width scoremat qseq best seqDB =
  Util.fill_matrix wgts 0.0;
  let nbest = Array.length best in
  for i=0 to nbest-1 do
    let _, hit = best.(i) in
    let dbseq = Hashtbl.find seqDB hit.dbkey in
    Scoring.setup_scoring_mat ~reset:true scoremat qseq dbseq;
(*     Util.write_matrix ~fname:("score."^(string_of_int i)^"mat") "%8.2f " scoremat; *)
    for j=0 to hit.alen-1 do
      match hit.qal.(j) with
	Some(qdx) ->
	  let start = max (j-width) 0 in
	  let stop = min (j+width) hit.alen in
	  let s = Align.score_alignment hit.qal hit.dal scoremat start stop in
(* 	  printf "%d %d %d %d %d %f\n" j qdx start stop len s; *)
	  wgts.(qdx).(i) <- s
      | None -> ()
    done;
  done;
(*   Util.write_matrix ~fname:("X.wgt") ~m:(qseq.seqlen) ~n:(nbest) "%f " wgts; *)
  Array.iter (shift_to_min 1.0) wgts;
  Array.iter normalize wgts;;

