(** Main entry point for counting columns, conservation *)

open Printf
open Align
open Sequence
open Util
open Scoring
open Matrices
open Query_align
open Respair

let topk = ref 10				 (* Number of hits to consider *)
let query_keys = ref [||]			 (* Keys of queries *)
let name = ref ""
let neighidxs = ref [||]

let setup_neighidxs s =
  neighidxs := Array.map int_of_string (Util.parse_strings s)

(* let setup_neighidxs s = *)
(*   let sidxs = Util.parse_strings s in *)
(*   let iter s = Hashtbl.add neighidxs (int_of_string s) () in *)
(*   Array.iter iter sidxs *)

open Ckarg  
(** Options for program execution *)
let prog_opts = [
  "-topk","int","Max number of alignments for each query. May be less than original topk but not more.",
    Set_int(topk);
  "-queries","'key1 [key2 ...]'",
    "Select the queries to search for in the DB.  default = whole sequence database",
    String(fun s->query_keys:=Util.parse_strings s);
  "-dir","directory","Add the directory to search path for data files",String(Find.add_dir);
  "-name","string","String to append to output files",Set_string(name);
  "-neighidxs","'n1 [n2 ...]'","List of neighbor indices to use for prediction",
     String(setup_neighidxs)
]

let options = prog_opts @ Seqdb.options @ Weighting.options

let usg = 
  sprintf "usage: %s [options]  -{filetype} seqfile -index indexfile [-{filetype} seqfile2 ...]"
    (Filename.basename Sys.executable_name)
let deffn s = failwith ("All arguments must have an option, "^s^" does not")

(* MAIN *)
let _ = begin
  parse options deffn usg;
  let seqDB = Seqdb.seqDB in
  let nDB = Hashtbl.length seqDB in

  (* Setup array for list of neighbors *)
  let topk = !topk in
  let neighidxs =
    if Array.length !neighidxs = 0 
    then Array.init topk (fun i-> i)
    else !neighidxs
  in
  let nneighs = Array.length neighidxs in

  (* Verify the sequence DB *)
  ignore(Seqdb.verify_db seqDB);

  let query_keys = !query_keys in
  let nq = Array.length query_keys in
  printf "%d DB entries imported, %d queries selected\n" nDB nq;
  flush_all ();

  (* Add default scoring *)
  Scoring.default_score (seq_gsm_scores ~scale:(!Scoring.wgt) !Matrices.gmat);
  
  let maxlen = Seqdb.maxlen_seqDB seqDB in
  let scoremat = Array.make_matrix maxlen maxlen 0.0 in
  let weights = Array.make_matrix maxlen nneighs 1.0 in


  let totaltime = ref 0.0 in
  (* Loop over all queries *)
  for i=0 to nq-1 do
    let _ = Timing.interval () in
    let qkey = query_keys.(i) in
    printf "%4d: %s " i qkey; flush stdout;
    let best =
      let f = open_in (Find.file (qkey ^ ".topk")) in
      let b : (float * topk_hit) array = Pervasives.input_value f in
      close_in f;
      b
    in
    let qseq = Hashtbl.find seqDB qkey in
    let qlen = qseq.seqlen in

    let qpairs : respair array = 
      let f = open_in (Find.file (qkey ^ ".psnn")) in
      let p = Pervasives.input_value f in
      close_in f;
      p
    in
    let qpairmatrix = Respair.pairmatrix qpairs qlen in
    
    (* A 2D array of stats for each type of mm to ss *)
    let qpairstats = Array.make_matrix qlen qlen [||] in
    for i=0 to qlen-1 do
      for j=i+1 to qlen-1 do
	qpairstats.(i).(j) <- Array.init (ss+1) (fun _-> Avgstd.make ());
      done;
    done;

    (* How many hits to really check against *)
    let nbest = Array.length best in

(*     let truek = *)
(*       if nbest < topk then *)
(* 	eprintf "Warning: %s has only %d top-k entries, truncating from %d\n%!" qkey nbest topk; *)
(*       min topk nbest *)
(*     in *)
(*     let best = Array.sub best 0 truek in *)

    (* Fill in the weights *)
    (!Weighting.scheme) weights scoremat qseq best seqDB;
    Array.iter Weighting.normalize weights;

    (* Iterate over all hits *)
    for ni=0 to nneighs-1 do
      if neighidxs.(ni) >= nbest then 
	eprintf "Warning: %s has only %d top-k entries, skipping requested neighbor %d\n" 
	  qkey nbest neighidxs.(ni)
      else begin
	let hi = neighidxs.(ni) in
	let _,h = best.(hi) in
	let dkey, alen, qal, dal = h.dbkey, h.alen, h.qal, h.dal in

	let dseq = Hashtbl.find seqDB dkey in
	let dlen = dseq.seqlen in
	let dpairs : respair array = 
	  let f = open_in (Find.file (dkey ^ ".psnn")) in
	  let p = Pervasives.input_value f in
	  close_in f;
	  p
	in
	let dpairmatrix = Respair.pairmatrix dpairs dlen in

	(* Look at each pair of aligned residues *)
	for ai=0 to alen-1 do
	  for aj=ai+1 to alen-1 do
	    match qal.(ai), qal.(aj), dal.(ai), dal.(aj) with
	      Some(qi), Some(qj), Some(di), Some(dj) ->
		let qpair = qpairmatrix.(qi).(qj) in
		let dpair = dpairmatrix.(di).(dj) in
		let qstats = qpairstats.(qi).(qj) in
		for typ=mm to ss do
		  if qpair.idxs.(typ) <> -1 && dpair.idxs.(typ) <> -1 then begin
		    let w = (weights.(qi).(ni) +. weights.(qj).(ni)) *. 0.5 in
		    Avgstd.update qstats.(typ) ~w dpair.objs.(typ);
		  end
		done;
	    | _ -> ()
 	  done;
	done;
      end;
    done;

    (* Objective vector *)
    let predobj =
      let maxobj = Array.fold_left (fun m p -> Array.fold_left max m p.idxs) 0 qpairs in
      Array.make (maxobj+1) nan
    in
    
    (* Indexes that are good *)
    let indexes = Array.make (Array.length predobj) false in
    
    (* Fill objective vector *)
    for i=0 to qlen-1 do
      for j=i+1 to qlen-1 do
	for typ=mm to ss do
	  let avg,_,cnt = Avgstd.finalize qpairstats.(i).(j).(typ) in
	  if cnt > 0.0 then begin
	    let objidx = qpairmatrix.(i).(j).idxs.(typ) in
	    predobj.(objidx) <- avg;
(* 	    trueobj.(objidx) <- qpairmatrix.(i).(j).objs.(typ); *)
	    indexes.(objidx) <- true
	  end;
	done;
      done;
    done;

    let f = open_out (qkey^(!name)^".pred") in
    Array.iter (fprintf f "%f\n") predobj;
    close_out f;

(*     let f = open_out (qkey^(!name)^".idx") in *)
(*     Array.iter (fun b-> let i=if b then 1 else 0 in fprintf f "%d\n" i) indexes; *)
(*     close_out f; *)

    (* Debugging *)
    let dbg = Array.map (fun _-> (0.0,0.0,0,0,0)) predobj in
    for i=0 to (Array.length qpairs)-1 do
      for typ=mm to ss do
	let objidx = qpairs.(i).idxs.(typ) in
	let trueobj = qpairs.(i).objs.(typ) in
	if objidx >= 0 then
	  dbg.(objidx) <- (trueobj, predobj.(objidx), qpairs.(i).pos1, qpairs.(i).pos2, typ);
      done;
    done;

(*     let f = open_out (qkey^(!name)^".dbg") in *)
(*     for i=0 to (Array.length dbg)-1 do *)
(*       let x,y,ii,jj,tt = dbg.(i) in *)
(*       fprintf f "%f %f %d %d %d\n" x y ii jj tt; *)
(*     done; *)
(*     close_out f; *)

    let f = open_out (qkey^(!name)^".dist") in
    for i=0 to (Array.length dbg)-1 do
      let x,_,_,_,_ = dbg.(i) in
      fprintf f "%f\n" x;
    done;
    close_out f;
    
    let t = Timing.interval () in
    printf "%.2f\n" t;
    totaltime := !totaltime +. t;
  done;
  printf "Totaltime: %.2f\n" !totaltime;
end

