(** Main program to print alignments for topk hits for database
    search *)

open Printf
open Align
open Sequence
open Util
open Scoring
open Matrices
open Align_io
open Query_align


let outformat = ref Standard		(* Format of output *)
let trans_table = ref "01" 
let outseq = ref AAsequence		(* Type of sequence to output in Standard *)
let output_scoremat = ref false		(* Whether to store the scoring matrix *)

let idxname = Sequence.idxname

open Ckarg
let prog_opts = [
  "-o","key","Output format",SymbolArb(true,[
      "standard","Standard format, includes score",(fun ()->outformat:=Standard);
      "pir","PIR sequence format",(fun ()->outformat:=PIR);
    ]);
  "-outseq","key","Style of output format",SymbolArb(false,[
      "aa","Amino acid sequence",(fun ()->outseq:=AAsequence);
      "ss","Secondary structure sequence",(fun ()->outseq:=SSsequence);
      "sspred2true","Secondary structure sequence",(fun ()->outseq:=SSpred2true);
      "index","Last index provided, must set the translation table using -trans",(fun ()->outseq:=IndexSeq(!idxname));
    ]);
  "-idxname","string","Set the name of the next index",
    Set_string(idxname);
  "-trans","string","Set the translation table to be used for an index",Set_string(trans_table);
  "-output_scoremat","","Leave the computed DP scoring matrix in file 'scores' for debuggin purposes",Set(output_scoremat);
]

let infile = ref ""
let deffn s = infile := s

let options = prog_opts @ Seqdb.options @ Scoring.options
let usg = 
  sprintf "usage: %s [options] -{filetype} query.topk\nMake sure to include the original sequences, usually with -sdb"
  (Filename.basename Sys.executable_name)


let _ = begin
  parse options deffn usg;
  Scoring.default_score (seq_gsm_scores ~scale:(!Scoring.wgt) !Matrices.gmat);
  let seqDB = Seqdb.seqDB in

  let infile = !infile in
  let qkey = Util.basename infile in
  let qseq = 
    try Hashtbl.find seqDB qkey 
    with _ -> failwith ("Could not find query "^qkey^" in the database")
  in
  let f = open_in infile in
  let best : (float * topk_hit) array = Pervasives.input_value f in
  close_in f;
  for i=0 to (Array.length best)-1 do
    let score,hit = best.(i) in
    let dkey = hit.dbkey in
    let dseq = 
      try Hashtbl.find seqDB dkey
      with _ -> failwith ("Could not find query "^dkey^" in the database")
    in
    let seqtrans = Align_io.get_outseq_array qseq dseq !outseq in
    print_alignment !outformat ~ids:(qkey,dkey) seqtrans hit.qal hit.dal score hit.alen;
    if !output_scoremat then begin
      let m,n = qseq.seqlen, dseq.seqlen in
      let smat = Array.make_matrix m n 0.0 in
      Scoring.setup_scoring_mat ~reset:false smat qseq dseq;
      let fname = sprintf "%s-%s.smat" qkey dkey in
      Util.write_matrix ~fname:fname "%8.3f " smat;
    end;
  done;
end
