(** Module for database operations on sequences *)

open Printf
open Sequence

type seqdb = (string, Sequence.sequence) Hashtbl.t

(** Module level sequence database *)
let seqDB : seqdb = Hashtbl.create 1000
(** Module level threshold for determining pos/neg class*)
let thresh = ref 0.0

(** Read a collection of sequence-related files out a list of files *)
let read_seqs ?(thresh= !thresh) ?(seqDB=seqDB) filet fname  =
  let files = Array.of_list (Util.read_strings fname) in
  let n = Array.length files in
  for i=0 to n-1 do
    let key = Util.basename files.(i) in
    let s = 
      if Hashtbl.mem seqDB key
      then Hashtbl.find seqDB key
      else Sequence.new_seq () 
    in
    let ss = Sequence.read_seq filet ~thresh s files.(i) in
    Hashtbl.replace seqDB key ss;
  done;;

(** Read a binary sequence database, an array of type [sequence]
    written using [Pervasives.output_value] *)
let read_bin_seqDB fname =
  let f = open_in fname in
  let seqDB : seqdb = Pervasives.input_value f in
  close_in f;
  seqDB;;

(** Append one sdb to another, no error checking done *)
let append_seqDB fname = 
  let seqDB2 = read_bin_seqDB fname in
  Hashtbl.iter (fun k s->Hashtbl.replace seqDB k s) seqDB2

(** Find the maximum length of sequences in the sequence database *)
let maxlen_seqDB seqDB = 
  Hashtbl.fold (fun _ s x-> max x s.seqlen) seqDB 0;;

(** Verify the database by running verify_seq on each sequence *)
let verify_db seqDB =
  let badseqs = Queue.create () in
  let check key seq =
    try verify_seq key seq
    with Bad_sequence (a,b,c,d) -> Queue.add (a,b,c,d) badseqs
  in
  Hashtbl.iter check seqDB;
  badseqs

(** Print basic stats on the sequence database *)
let seqDB_report printall seqDB =
  let nDB = Hashtbl.length seqDB in
  let badseqs = verify_db seqDB in
  let lastkey = Hashtbl.fold (fun k _ _-> k) seqDB "" in

  let feat, index_names = 
    if nDB>0 then
      let s = Hashtbl.find seqDB lastkey in
      (feature_lengths s, Util.hash_keys s.indexes)
    else [], [] in
  printf "DBsize: %d\nFeatures: " nDB;
  List.iter (fun (s,n)->if n>0 then printf "%s " s) feat;
  List.iter (fun s -> printf "%s " s) index_names;
  printf "(%d indexes)\n" (List.length index_names);
  printf "Nbad seqs: %d\nBad seqs:\n" (Queue.length badseqs);
  let printbad (id, feat, flen, slen) =
    printf "%10s %10s: %5d != %5d\n" id feat flen slen in
  Queue.iter printbad badseqs;
  if printall then begin
    printf "--All sequences--\n%10s %6s\n" "ID" "Length";
    Hashtbl.iter (fun k s->printf "%10s %6d\n" k s.seqlen) seqDB;
  end;;



open Ckarg
let options = [
  "-sdb","file","Read a binary sequence database created with seqdbtool",
    String(append_seqDB);
  "-profile","file","Use PSI-blast profiles",String(read_seqs Profile);
  "-fasta","file","FASTA files",String(read_seqs FASTA);
  "-gkprofile","file","Profiles from PSI-Blast according the George's Modification",
    String(read_seqs GKprofile);
  "-ss","file","Read secondary structure from files in FASTA format, H:helix, E:strand, C:coil",
    String(read_seqs SSfasta);
  "-yasspp","file","YASSPP output file",String(read_seqs YASSPP);
  "-yasspptrue","file","YASSPP output file, use predicted as true secondary structure",
    String(read_seqs YASSPPtrue);
  "-sasa","file","SASA in space separated float format",String(read_seqs SASA);
  "-index","file","Read a vector of integers describing each AA of the sequence giving it current label",
    String(read_seqs Index);
  "-idxname","string","Set the name of the next index",
    Set_string(Sequence.idxname);
  "-t","float","Set the weight of thresholding for -thresh and -thresh2true",
    (Set_float thresh);
  "-thresh","file",
    "Read floating point values, below t becomes 0, above becomes 1 in an index with current\
     label. Useful for prediction inputs.",
    String(read_seqs Threshold);
  "-tclass","file","Read an index and set up a true class vector for it",
    String(read_seqs TClass);
  "-p2class","file",
    "Read probabilities of being in positive class, set up predicted class vector",
    String(read_seqs P2Class);
]
